<?php

namespace Drupal\weather\Form;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Database\Database;
use Drupal\Core\Entity\ContentEntityForm;
use Drupal\Core\Entity\EntityManagerInterface;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\weather\Entity\WeatherDisplayInterface;
use Drupal\weather\Service\HelperService;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form controller for the weather_display entity edit forms.
 *
 * @ingroup weather_display
 */
class WeatherDisplayForm extends ContentEntityForm {

  /**
   * Weather helper service.
   *
   * @var \Drupal\weather\Service\HelperService
   */
  protected $weatherHelperService;

  /**
   * Weather display storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $weatherDisplayStorage;

  /**
   * WeatherDisplayForm constructor.
   *
   * @param \Drupal\Core\Entity\EntityRepositoryInterface $entity_repository
   *   The entity repository service.
   * @param \Drupal\weather\Service\HelperService $weatherHelperService
   *   Weather helper service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   Entity Type manager service.
   * @param \Drupal\Core\Entity\EntityTypeBundleInfoInterface|null $entity_type_bundle_info
   *   The entity type bundle service.
   * @param \Drupal\Component\Datetime\TimeInterface|null $time
   *   The time service.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function __construct(EntityRepositoryInterface $entity_repository, HelperService $weatherHelperService, EntityTypeManagerInterface $entityTypeManager, EntityTypeBundleInfoInterface $entity_type_bundle_info = NULL, TimeInterface $time = NULL) {
    parent::__construct($entity_repository, $entity_type_bundle_info, $time);

    $this->weatherHelperService = $weatherHelperService;
    $this->weatherDisplayStorage = $entityTypeManager->getStorage('weather_display');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity.repository'),
      $container->get('weather.helper'),
      $container->get('entity_type.manager'),
      $container->get('entity_type.bundle.info'),
      $container->get('datetime.time')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildForm($form, $form_state);
    $displayNumber = NULL;
    $savedConfig = [];

    switch ($this->operation) {
      case 'add':
        $display_type = WeatherDisplayInterface::SYSTEM_WIDE_TYPE;
        break;

      case 'edit':
        $display_type = WeatherDisplayInterface::SYSTEM_WIDE_TYPE;
        $savedConfig = $this->entity->config->getValue()[0];
        break;

      case 'add_user':
        // This is not implemented for now.
        $display_type = WeatherDisplayInterface::USER_TYPE;
        break;

      case 'default':
        $display_type = WeatherDisplayInterface::DEFAULT_TYPE;
        $displayNumber = 1;
        break;

      default:
        $display_type = NULL;
    }

    if (!$this->entity->isNew()) {
      $displayNumber = $this->entity->number->value;
    }
    $defaultConfig = $this->weatherHelperService->getDisplayConfig(WeatherDisplayInterface::DEFAULT_TYPE);

    $form['config'] = [
      '#type' => 'fieldset',
      '#title' => t('Display configuration'),
      '#description' => t('Customize the weather display.'),
      '#collapsible' => TRUE,
      '#collapsed' => FALSE,
      '#tree' => TRUE,
    ];
    $form['config']['temperature'] = [
      '#type' => 'select',
      '#title' => t('Temperature'),
      '#description' => t('Unit for displaying temperatures.'),
      '#default_value' => $savedConfig['temperature'] ?? $defaultConfig['temperature'],
      '#options' => [
        'celsius' => t('Celsius'),
        'fahrenheit' => t('Fahrenheit'),
        'celsiusfahrenheit' => t('Celsius / Fahrenheit'),
        'fahrenheitcelsius' => t('Fahrenheit / Celsius'),
      ],
    ];
    $form['config']['windspeed'] = [
      '#type' => 'select',
      '#title' => t('Wind speed'),
      '#description' => t('Unit for displaying wind speeds.'),
      '#default_value' => $savedConfig['windspeed'] ?? $defaultConfig['windspeed'],
      '#options' => [
        'kmh' => t('km/h'),
        'mph' => t('mph'),
        'knots' => t('Knots'),
        'mps' => t('meter/s'),
        'beaufort' => t('Beaufort'),
      ],
    ];
    $form['config']['pressure'] = [
      '#type' => 'select',
      '#title' => t('Pressure'),
      '#description' => t('Unit for displaying pressure.'),
      '#default_value' => $savedConfig['pressure'] ?? $defaultConfig['pressure'],
      '#options' => [
        'hpa' => t('hPa'),
        'kpa' => t('kPa'),
        'inhg' => t('inHg'),
        'mmhg' => t('mmHg'),
      ],
    ];
    $form['config']['distance'] = [
      '#type' => 'select',
      '#title' => t('Distance'),
      '#description' => t('Unit for displaying distances.'),
      '#default_value' => $savedConfig['distance'] ?? $defaultConfig['distance'],
      '#options' => ['kilometers' => t('Kilometers'), 'miles' => t('UK miles')],
    ];
    $form['config']['show_sunrise_sunset'] = [
      '#type' => 'checkbox',
      '#title' => t('Show times of sunrise and sunset'),
      '#default_value' => $savedConfig['show_sunrise_sunset'] ?? $defaultConfig['show_sunrise_sunset'],
      '#description' => t('Displays the times of sunrise and sunset. This is always the local time.'),
    ];
    $form['config']['show_windchill_temperature'] = [
      '#type' => 'checkbox',
      '#title' => t('Show windchill temperature'),
      '#default_value' => $savedConfig['show_windchill_temperature'] ?? $defaultConfig['show_windchill_temperature'],
      '#description' => t('Displays the windchill temperature. This is how the temperature <q>feels like</q>. Note that windchill temperature is only defined for temperatures below 10 °C (50 °F) and wind speeds above 1.34 m/s (3 mph).'),
    ];
    $form['config']['show_abbreviated_directions'] = [
      '#type' => 'checkbox',
      '#title' => t('Show abbreviated wind directions'),
      '#default_value' => $savedConfig['show_abbreviated_directions'] ?? $defaultConfig['show_abbreviated_directions'],
      '#description' => t('Displays abbreviated wind directions like N, SE, or W instead of North, Southeast, or West.'),
    ];
    $form['config']['show_directions_degree'] = [
      '#type' => 'checkbox',
      '#title' => t('Show degrees of wind directions'),
      '#default_value' => $savedConfig['show_directions_degree'] ?? $defaultConfig['show_directions_degree'],
      '#description' => t('Displays the degrees of wind directions, for example, North (20°).'),
    ];
    $form['type'] = [
      '#type' => 'value',
      '#value' => $display_type,
    ];
    $form['number'] = [
      '#type' => 'value',
      '#value' => $displayNumber,
    ];

    // Show a 'reset' button if editing the default display.
    if ($this->operation == 'default') {
      $form['actions']['reset'] = [
        '#type' => 'submit',
        '#value' => $this->t('Reset'),
        '#weight' => 10,
        '#submit' => ['::submitForm', '::save'],
      ];
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $triggeredBy = $form_state->getTriggeringElement();
    if ($triggeredBy && $triggeredBy['#id'] == 'edit-reset') {
      $defaultConfig = $this->weatherHelperService->getDefaultConfig();
      $form_state->setValue('config', $defaultConfig);
    }

    parent::submitForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    // Set display number before save.
    if ($form_state->getValue('number') == NULL) {
      $type = $form_state->getValue('type');
      $free_number = $this->getFreeDisplayNumber($type);
      $this->entity->set('number', $free_number);
    }

    $this->entity->set('config', $form_state->getValue('config'));

    // Make sure we have only one instance of display with type 'default'.
    if ($this->operation == 'default') {
      $defaultDisplay = $this->weatherDisplayStorage->loadByProperties([
        'type' => WeatherDisplayInterface::DEFAULT_TYPE,
      ]);
      $defaultDisplay = reset($defaultDisplay);
      if ($defaultDisplay instanceof WeatherDisplayInterface) {
        $this->entity->id = $defaultDisplay->id();
        $this->entity->enforceIsNew(FALSE);
      }
    }

    $status = parent::save($form, $form_state);
    if ($status == SAVED_NEW) {
      $message = $this->t('Created new Weather display');
    }
    else {
      $message = $this->t('Updated existing Weather display');
    }

    $this->messenger()->addStatus($message);
    $form_state->setRedirectUrl(Url::fromRoute('weather.settings'));

    return $status;
  }

  /**
   * Finds first free display number for given display type.
   */
  protected function getFreeDisplayNumber(string $displayType) : int {
    $used_numbers = Database::getConnection()
      ->select('weather_display', 'wd')
      ->fields('wd', ['number'])
      ->condition('type', $displayType)
      ->execute();

    $free_number = 1;
    foreach ($used_numbers as $row) {
      if ($row->number > $free_number) {
        break;
      }
      else {
        $free_number++;
      }
    }
    return $free_number;
  }

}
